using LinearAlgebra
using Random: MersenneTwister
using Test

using BenchmarkTools

using DualTVDD:
    DualTVL1ROFOpProblem, ProjGradAlgorithm, ChambolleAlgorithm, DualTVDDAlgorithm, DualTVDDSurrogateAlgorithm,
    init, step!, fetch_u, normB

const seed = 47

@testset "B = I" begin
    g = Float64[0 2; 1 0]
    prob = DualTVL1ROFOpProblem(g, I, 1e-10)

    @testset for alg in (ProjGradAlgorithm(prob, τ=1/8), ChambolleAlgorithm(prob))
        ctx = init(alg)
        #@test 0 == @ballocated step!($ctx)
        for i in 1:100
            step!(ctx)
        end
        u = fetch_u(ctx)
        @test u ≈ g
    end
end

@testset "B = rand(...)" begin
    g = Float64[0 2; 1 0]
    rng = MersenneTwister(seed)
    B = rand(rng, length(g), length(g))
    prob = DualTVL1ROFOpProblem(g, B, 1e-10)

    @testset for alg in (ProjGradAlgorithm(prob, τ=1/8), ChambolleAlgorithm(prob))
        ctx = init(alg)
        #@test 0 == @ballocated step!($ctx)
        for i in 1:100
            step!(ctx)
        end
        u = fetch_u(ctx)
        @test vec(u) ≈ B * vec(g)
    end
end

@testset "DualTVDDAlgorithm" begin
    n = 5
    ninner = 100
    rng = MersenneTwister(seed)
    g = rand(rng, n, n)
    B = Diagonal(rand(rng, n^2))
    # big λ is ok, since we test for inter-subdomain communication
    prob = DualTVL1ROFOpProblem(g, B, 100.)

    algref = ChambolleAlgorithm(prob)
    alg = DualTVDDAlgorithm(prob; M=(2,2), overlap=(2,2), ninner, parallel = false, σ = 0.25, subalg = x -> ChambolleAlgorithm(x))

    stref = init(algref)
    st = init(alg)
    #@test 0 == @ballocated step!($ctx)
    for i in 1:1000*ninner
        step!(stref)
    end
    for i in 1:1000
        step!(st)
    end
    @test fetch_u(st) ≈ fetch_u(stref)
end

@testset "DualTVDDSurrogateAlgorithm" begin
    d = 2
    n = 5
    ninner = 100
    rng = MersenneTwister(seed)
    g = rand(rng, n, n)
    B = rand(rng, n^2, n^2)
    B = I + 0.5 .* (B + B') ./ n^2
    M = ntuple(_->2, d)
    overlap = ntuple(_->2, d)
    parallel = true
    σ = 0.25
    # big λ is ok, since we test for inter-subdomain communication
    prob = DualTVL1ROFOpProblem(g, B, 1000.)

    algref = ChambolleAlgorithm(prob)
    #algref = DualTVDDAlgorithm(prob; M, overlap, ninner=1, parallel, σ,
    #    subalg = x -> SurrogateAlgorithm(x; τ=2*normB(prob), ninner))
    alg = DualTVDDSurrogateAlgorithm(prob; M, overlap, ninner, parallel, σ,
        subalg = x -> ChambolleAlgorithm(x))

    stref = init(algref)
    st = init(alg)
    #@test 0 == @ballocated step!($ctx)
    for i in 1:1000
        step!(stref)
    end
    for i in 1:1000
        step!(st)
    end
    @test fetch_u(st) ≈ fetch_u(stref)
end
